"""Package containig various model files for fitting.

This package contains the various MRI models currently implemented for the
toolbox. In additon, the "GeneralModel" can be run with a simple text file
to devine new model. An exemplary textfile for simple models can be
generated by running the "genDefaultModelfile" function of the "GeneralModel".
"""
# __all__ = ["BiExpDecay",
#            "DiffdirLL",
#            "GeneralModel",
#            "ImageReco",
#            "IRLL",
#            "template",
#            "VFA"]
from .BiExpDecay import Model as BiExpDecay
from .DiffdirLL import Model as DiffdirLL
from .GeneralModel import Model as GeneralModel
from .ImageReco import Model as ImageReco
from .IRLL import Model as IRLL
from .VFA import Model as VFA
